/*
 * interdiff - create incremental patch between two against a common source
 *
 * Utility functions.
 * Copyright (C) 2001  Marko Kreen
 * Copyright (C) 2001  Tim Waugh <twaugh@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <fnmatch.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "util.h"

/* safe malloc */
void *xmalloc (size_t size)
{
	void *res = malloc(size);
	if (!res) {
		fprintf (stderr, "No memory available\n");
		exit(1);
	}
	return res;
}

/* safe strdup */
char *xstrdup (const char *s)
{
	size_t len = strlen (s) + 1;
	char *result = xmalloc (len);
	memcpy (result, s, len);
	return result;
}

/* only copy the first n characters of s */
char *xstrndup (const char *s, const size_t n)
{
	char *result;
	result = xmalloc(n + 1);
	strncpy(result, s, n);
	result[n] = '\0';
	return result;
}

/*
 * Pattern operations.
 */
/* list of patterns */

struct patlist {
	struct patlist *next;
	char *pattern;
};

void patlist_add(struct patlist **dst, const char *s)
{
	struct patlist *item = xmalloc (sizeof *item);
	item->pattern = xstrdup (s);
	item->next = *dst;
	*dst = item;
}

int patlist_match(struct patlist *list, const char *s)
{
	while (list) {
		if (!fnmatch (list->pattern, s, 0))
			return 1;
		list = list->next;
	}
	return 0;
}

void patlist_free(struct patlist **list)
{
	struct patlist *l, *next;
	for (l = *list; l; l = next) {
		next = l->next;
		free (l->pattern);
		free (l);
	}
	*list = NULL;
}


char *progname = "(null)";
void set_progname(const char *s)
{
	progname = xstrdup(s);
}


FILE *xopen (const char *name, const char *mode)
{
	FILE *f;

	f = fopen(name, mode);
	if (!f) {
		perror(name);
		exit(1);
	}
	return f;
}

FILE *xopen_seekable (const char *name, const char *mode)
{
	/* If it's seekable, good.  If not, we need to make it seekable. */
	FILE *f = xopen (name, mode);

	if (fseek (f, 0L, SEEK_SET) != 0) {
		const size_t buflen = 64 * 1024;
		char *buffer = xmalloc (buflen);
		FILE *tmp = tmpfile();
		if (!tmp) {
			perror("tmpfile()");
			exit(1);
		}

		while (!feof (f)) {
			size_t count = fread (buffer, 1, buflen, f);
			if (count < 1)
				break;
			fwrite (buffer, count, 1, tmp);
		}

		free (buffer);
		fclose (f);

		f = tmp;
		fseek (f, 0L, SEEK_SET);
	}

	return f;
}

/* safe pipe/popen.  mode is either "r" or "w" */
FILE * xpipe(const char * cmd, pid_t *pid, const char *mode, ...)
{
	va_list ap;
	int fildes[2];
	int child;
	int nargs = 0;
	char *argv[128], *arg;
	FILE *res;
	
	if (!mode || (*mode != 'r' && *mode != 'w')) {
		fprintf(stderr, "xpipe: bad mode: %s\n", mode);
		exit(1);
	}
	
	va_start(ap, mode);
	do {
		arg = va_arg(ap, char *);
		argv[nargs++] = arg;
		if (nargs >= 128) {
			fprintf(stderr, "xpipe: too many args\n");
			exit(1);
		}
	} while (arg != NULL);
	va_end(ap);
	
	fflush (NULL);
	pipe (fildes);
	child = fork ();
	if (child == -1) {
		perror ("fork");
		exit (1);
	}
	if (child == 0) {
		if (*mode == 'r') {
			close (fildes[0]);
			close (1);
			dup (fildes[1]);
			close (fildes[1]);
		} else {
			close (fildes[1]);
			close (1);
			dup(2);
			close (0);
			dup (fildes[0]);
			close (fildes[0]);
		}
		execvp (cmd, argv);
		/* Shouldn't get here */
		perror("execvp");
		exit (1);
	}
	if (pid != NULL)
		*pid = child;
	
	if (*mode == 'r') {
		close (fildes[1]);
		res = fdopen (fildes[0], "r");
	} else {
		close (fildes[0]);
		res = fdopen (fildes[1], "w");
	}
	if (res == NULL) {
		perror("fdopen");
		exit(1);
	}
	return res;
}

/*
 * stuff needed for non-GNU systems
 */

#ifndef HAVE_GETLINE
#define GLSTEP 512

/* suboptimal implementation of glibc's getline() */
int getline(char **line, size_t *n, FILE *f)
{
	char *p;
	int len;
	
	if (*line == NULL || *n < 2) {
		p = realloc(*line, GLSTEP);
		if (!p)
			return -1;
		*line = p;
		*n = GLSTEP;
	}
	
	p = fgets(*line, *n, f);
	if (!p)
		return -1;
	
	len = strlen(p);
	while (p[len - 1] != '\n') {
		p = realloc(*line, *n + GLSTEP);
		if (!p)
			return -1;
		*line = p;
		*n += GLSTEP;
		
		p = fgets(p + len, *n - len, f);
		if (!p)
			break;

		len = len + strlen(p);
	}
	return len;
}

#endif

